import os
import logging
import bzr
import phase
import shutil
import shlex
import re
import hashlib
import sys
from .contain import (
    LXCContainer,
    NoSuchContainerError,
)
from .utils import (
    needs_root,
    mkdirs,
    mksymlink,
)

MOJO_ROOT = "/srv/mojo"


#######################################
# Spec
#######################################

class SpecUninitializedError(Exception):
    pass


class ConfigNotFoundException(Exception):
    def __init__(self, message, search_list):
        self.message = message
        self.search_list = search_list


class Spec(object):
    def __init__(self, local_dir, remote=None):
        self.local_dir = local_dir
        self.local_branch = bzr.BzrBranch(self.local_dir)
        if remote:
            self.remote_branch = bzr.BzrBranch(remote)
        elif self.is_initialized:
            if self.local_branch.parent:
                self.remote_branch = bzr.BzrBranch(self.local_branch.parent)
            else:
                self.remote_branch = None
        else:
            raise ValueError("No remote or local branch specified for Spec")
        self._manifest = None

    @property
    def is_initialized(self):
        return self.local_branch is not None and self.local_branch.is_branch

    def initialize(self):
        "Download or update spec on disk"
        if not os.path.isdir(self.local_dir):
            os.makedirs(self.local_dir)
        if self.local_branch.is_branch:
            self.local_branch.pull(src_location=self.remote_branch.location,
                                   overwrite=True)
        else:
            self.local_branch = self.remote_branch.clone(self.local_dir)
        self.local_branch.revert()
        self.local_branch.cleantree()

    @property
    def manifest(self):
        logging.info("Retreive the Spec's manifest")
        if self._manifest:
            return self._manifest
        if not self.is_initialized:
            return None
        self._manifest = Manifest(self)
        return self._manifest

    def get_config(self, base, stage=None):
        """Locate an stage-qualified configuration file.
           Configs are search in this order:
             * <stage>/<base>
             * <base>-<stage>
             * <base>
           The first one found is returned."""
        if not self.is_initialized:
            raise SpecUninitializedError("Spec has no local copy: {}"
                                         "".format(self.local_dir))
        config_base = os.path.join(self.local_dir, base)
        config_names = [config_base]
        if stage:
            config_names.insert(0, "{}-{}".format(config_base, stage))
            config_names.insert(0, os.path.join(self.local_dir, stage, base))
        for config in config_names:
            if os.path.exists(config):
                return config
        raise ConfigNotFoundException("Config not found. Is MOJO_ENV_TAG set "
                                      "or use the --stage option",
                                      config_names)


#######################################
# Manifest
#######################################

class InvalidManifestException(Exception):
    pass


class Manifest(object):
    required_commands = ('collect', 'repo', 'deploy')

    def __init__(self, spec):
        super(Manifest, self).__init__()
        self.spec = spec
        self.config = spec.get_config('manifest')
        self.phases = []
        self.load()

    def load(self):
        commands = set()
        with open(self.config) as f:
            for line in f:
                bareline = line.strip()
                if bareline and not bareline.startswith("#"):
                    words = shlex.split(bareline)
                    command = words[0]
                    kwargs = {}
                    for arg in words[1:]:
                        k, v = arg.split("=", 1)
                        kwargs[k] = v
                    commands.add(command)
                    try:
                        self.phases.append(phase.build(command, **kwargs))
                    except phase.NoSuchPhaseError as e:
                        raise InvalidManifestException(e.message)
        for command in self.required_commands:
            if command not in commands:
                raise InvalidManifestException("Missing required phase: {}"
                                               "".format(command))

    def run(self, project, workspace, series, stage=None):
        "Run all phases defined in this manifest"
        for phase in self.phases:
            phase.run(project, workspace, series, stage)


#######################################
# Monitor
#######################################

class Monitor(object):

    def __init__(self):
        self.CHANGED = 100

    def run(self, project, workspace, series, stage=None):
        """"Check for new revnos for branches
            Store the previous revnos in monitor dir
        """
        self.spec = workspace.spec
        self.monitor_dir = os.path.join(project.root, 'monitor')
        mkdirs(self.monitor_dir)
        previous_revno = None
        new = False
        with open(self.spec.get_config('collect', stage)) as config:
            for line in config:
                if not line.strip().startswith("#") and not len(line.strip()) == 0:
                    options = {}
                    source = line.split()[1]
                    if ";" in source:
                        source, opt_string = source.strip().rsplit(";", 1)
                        for opt in opt_string.split(","):
                            options.update(dict((opt.split("="),)))
                    digest = hashlib.sha1(source).hexdigest()
                    branch = bzr.BzrBranch(source)
                    # Not a branch, could be a directory (@) or http source:
                    # ignore
                    if branch.is_branch:
                        current_revno = branch.revno
                        digestpath = os.path.join(self.monitor_dir, digest)
                        if os.path.exists(digestpath):
                            with open(digestpath) as digestfile:
                                for line in digestfile:
                                    previous_revno = line.strip()
                        else:
                            with open(digestpath, 'w') as digestfile:
                                digestfile.write(digest)
                        if not previous_revno == str(current_revno):
                            if "revno" in options:
                                logging.debug("Revno {} for branch {} has "
                                              "changed but configuration "
                                              "calls for revno {}"
                                              "".format(current_revno, source,
                                                        options["revno"]))
                                current_revno = options["revno"]
                            else:
                                new = True
                                logging.info("Revno {} for branch {} has "
                                             "changed from previous revno {}"
                                             "".format(current_revno, source,
                                                       previous_revno))
                            with open(digestpath, 'w') as digestfile:
                                digestfile.write(str(current_revno))
        if new:
            logging.info("Changes to branches for project: {}"
                         "".format(project.name))
            sys.exit(self.CHANGED)


#######################################
# Project
#######################################

class DuplicateProjectError(Exception):
    pass


class ProjectCreationError(Exception):
    pass


class Project(object):
    "A containerized mojo project"
    token_re = re.compile(r"[a-zA-Z0-9_.-]").search
    token_desc = "{} must only contain alphanumerics, underscores, or periods."

    def __init__(self, name, series, container_class=LXCContainer):
        if not name:
            raise ProjectCreationError("No project name specified. Use "
                                       "--project or set MOJO_PROJECT")
        if not bool(self.token_re(name)):
            raise ProjectCreationError(self.token_desc.format("Project names"))
        self.name = name
        self.series = series
        self.owner = os.environ.get('SUDO_USER')
        if not self.owner:
            self.owner = os.environ.get('USER', 'root')

        ### Paths ###
        # /srv/mojo/<prj>
        self.root = Project.root_path(self.name)
        # /srv/mojo/<prj>/<series>
        self.project_series = os.path.join(self.root, self.series)
        # /srv/mojo/<prj>/<series>/ROOTFS
        self.container_root = os.path.join(self.project_series, "ROOTFS")
        # /srv/mojo/<prj>/<series>/ROOTFS/srv/mojo/<prj>/<series>
        self.lxc_project_series = self.container_root + self.project_series

        self.container = container_class(self.name, self.series)

    @classmethod
    def root_path(cls, name):
        return os.path.join(MOJO_ROOT, name)

    @classmethod
    def list_series(cls, name):
        return [s for s in sorted(os.listdir(cls.root_path(name)))]

    @classmethod
    def list(cls, only_names):
        "List all Mojo projects"
        proj_list = []
        for proj in os.listdir(MOJO_ROOT):
            if proj in ['LOCAL', 'bin', 'lib']:
                continue
            if not only_names:
                series = " ".join(cls.list_series(proj))
                proj = "{}: {}".format(proj, series)
            proj_list.append(proj)
        return proj_list

    @needs_root
    def create(self):
        "Create a new LXC container and project"
        logging.info("Checking {} project path at {} ...".format(
            self.name, self.root))
        if os.path.exists(self.root):
            logging.warning("Project {} already exists at {}. Using existing "
                            "project".format(self.name, self.root))
        else:
            try:
                mkdirs(self.root, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project root: {}".format(e.strerror))

        if os.path.exists(self.project_series):
            logging.warning("Project {} project_series already exists at {}, "
                            "using it.".format(self.name, self.project_series))
        else:
            try:
                mkdirs(self.project_series, self.owner, mode=0700)
            except OSError as e:
                raise ProjectCreationError(
                    "Could not create project project_series: "
                    "{}".format(e.strerror))
        container_name = "{}.{}".format(self.name, self.series)
        lxc_root = "/var/lib/lxc/{}/rootfs".format(container_name)
        logging.info("Checking {} project LXC container at {} ..."
                     "".format(self.name, lxc_root))
        if not os.path.exists(lxc_root):
            logging.info("Not found, creating LXC container")
            # Can't use root location on precise's version of lxc
            # missing -B --dir option
            #self.container.create(self.container_root)
            self.container.create()

        logging.info("Checking {} project LXC container setup ..."
                     "".format(self.name))
        if not self.container.is_setup(lxc_root):
            logging.info("Not setup - setting up LXC container for mojo")
            self.container.setup()

        logging.info("Checking {} project container symlink {} -> {} ..."
                     "".format(self.name, lxc_root, self.container_root))
        if not os.path.lexists(self.container_root):
            os.symlink(lxc_root, self.container_root)

        logging.info("Checking {} project dir: {} ..."
                     "".format(self.name, self.lxc_project_series))
        if not os.path.exists(self.lxc_project_series):
            mkdirs(self.lxc_project_series, self.owner, mode=0700)

        sudoers_file = os.path.join(self.container_root, "etc", "sudoers.d",
                                    "ubuntu-all")
        logging.info("Writing {} sudoers file: {}".format(
                     self.name, sudoers_file))
        with open(sudoers_file, "w") as f:
            f.write("ubuntu ALL=(ALL) NOPASSWD: ALL\n")
        os.chmod(sudoers_file, 0400)

    def delete(self, force=True):
        """Delete a project and its LXC container - if no series remain,
        also delete project root"""
        try:
            shutil.rmtree(self.project_series)
        except OSError:
            if not force:
                raise
        try:
            self.container.destroy()
        except NoSuchContainerError:
            if not force:
                raise

        # If no series remain, delete project root
        if not Project.list_series(self.name):
            logging.info("Destroying project root: {}"
                         "".format(Project.root_path(self.name)))
            try:
                shutil.rmtree(Project.root_path(self.name))
            except OSError:
                if not force:
                    raise

    def ws_link(self, title):
        return os.path.join(self.project_series, str(title))

    def create_workspace(self, spec_url, title):
        ws_path = os.path.join(self.lxc_project_series, str(title))
        mkdirs(ws_path, self.owner)
        mksymlink(ws_path, self.ws_link(title))
        ws = Workspace(title, self.ws_link(title))
        ws.prepare()
        spec = Spec(ws.spec_dir, spec_url)
        spec.initialize()
        return ws

    def workspace(self, workspace_name):
        return Workspace(workspace_name, self.ws_link(workspace_name))

    def delete_workspace(self, workspace_name):
        ws = self.workspace(workspace_name)
        ws.delete()
        if os.path.lexists(self.ws_link(workspace_name)):
            os.unlink(self.ws_link(workspace_name))

    def workspace_list(self):
        return (w for w in os.listdir(self.lxc_project_series) if w != 'LOCAL')

    @property
    def utils(self):
        return self.container.utils


#######################################
# Workspace
#######################################

class NoSuchWorkspaceError(Exception):
    pass


class Workspace(object):
    "A mojo workspace"
    def __init__(self, name, path):
        self.name = name
        if not path:
            self.path = os.path.abspath(os.curdir)
            logging.info("No workspace path specified, using {}"
                         "".format(self.path))
        else:
            self.path = os.path.abspath(path)
        if not os.path.exists(self.path):
            raise NoSuchWorkspaceError(path)

    @property
    def spec(self):
        if os.path.exists(self.spec_dir):
            spec_repo = bzr.BzrBranch(self.spec_dir)
            if spec_repo.is_branch:
                return Spec(local_dir=self.spec_dir)
        return None

    @property
    def all_dirs(self):
        return [self.path,
                self.build_dir,
                self.local_dir,
                self.log_dir,
                self.spec_dir,
                ]

    def prepare(self):
        "Set up the workspace. Returns True on success, False on failure."
        for path in self.all_dirs:
            if not os.path.isdir(path):
                try:
                    os.makedirs(path)
                except OSError as e:
                    logging.error("{} ({})".format(e.strerror, e.filename))
                    return False
        return True

    def clean(self):
        "Reset the workspace"
        map(shutil.rmtree, (d for d in self.all_dirs if os.path.exists(d)))
        return self.prepare()

    def delete(self):
        "Destroy the workspace"
        realpath = os.path.realpath(self.path)
        shutil.rmtree(realpath)

    @property
    def build_dir(self):
        return os.path.join(self.path, "build")

    @property
    def repo_dir(self):
        return os.path.join(self.path, "charms")

    @property
    def local_dir(self):
        return os.path.join(self.path, "local")

    @property
    def log_dir(self):
        return os.path.join(self.path, "log")

    @property
    def spec_dir(self):
        return os.path.join(self.path, "spec")
