from unittest import TestCase
from test_bzr import (
    create_unique_branch,
)

from tempfile import mkdtemp
from shutil import rmtree
import commands
import os
import sys

# Make sure that the local mojo is importable
local_copy = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "..",
    "mojo")
if os.path.exists(local_copy) and os.path.isdir(local_copy):
    sys.path.insert(0, local_copy)
    print sys.path

try:
    import mojo
except ImportError:
    raise ImportError("Could not import local copy of mojo.")


class WorkspaceTest(TestCase):

    def setUp(self):
        super(WorkspaceTest, self).setUp()
        self.bad_path = "/INVALID/PATH"
        self.good_path = mkdtemp()
        self.addCleanup(rmtree, self.good_path)

        self.build_dir_name = "build"
        self.local_dir_name = "local"
        self.log_dir_name = "log"
        self.spec_dir_name = "spec"
        self.all_names = [
            self.build_dir_name,
            self.local_dir_name,
            self.log_dir_name,
            self.spec_dir_name,
        ]

    def test_workspace_sets_path(self):
        w = mojo.Workspace("workspace_name", self.good_path)
        self.assertEqual(w.path, self.good_path)

        with self.assertRaises(mojo.NoSuchWorkspaceError):
            w = mojo.Workspace("workspace_name", self.bad_path)

    def test_workspace_sets_default_path(self):
        os.chdir(self.good_path)
        w = mojo.Workspace("workspace_name", None)
        self.assertEqual(w.path, self.good_path)

    def test_prepares_path(self):
        w = mojo.Workspace("workspace_name", self.good_path)
        self.assertTrue(w.prepare())

        all_subdirs = [os.path.join(self.good_path, d) for d in self.all_names]
        for subdir in all_subdirs:
            self.assertTrue(os.path.exists(subdir))

    def test_prepares_new_path(self):
        w = mojo.Workspace("workspace_name", self.good_path)
        self.assertTrue(w.prepare())

        all_subdirs = [os.path.join(self.good_path, d) for d in self.all_names]
        for subdir in all_subdirs:
            self.assertTrue(os.path.exists(subdir))

    def test_prepare_errors_return_false(self):
        os.chmod(self.good_path, 0o555)
        w = mojo.Workspace("workspace_name", self.good_path)
        self.assertFalse(w.prepare())

    def test_returns_dirs(self):
        w = mojo.Workspace("workspace_name", self.good_path)

        build_dir = os.path.join(self.good_path, self.build_dir_name)
        self.assertEqual(w.build_dir, build_dir)

        local_dir = os.path.join(self.good_path, self.local_dir_name)
        self.assertEqual(w.local_dir, local_dir)

        log_dir = os.path.join(self.good_path, self.log_dir_name)
        self.assertEqual(w.log_dir, log_dir)

        spec_dir = os.path.join(self.good_path, self.spec_dir_name)
        self.assertEqual(w.spec_dir, spec_dir)


class SpecTest(TestCase):
    def setUp(self):
        super(SpecTest, self).setUp()

        self.bad_dir = mkdtemp()

        self.good_source = mkdtemp()
        self.addCleanup(rmtree, self.good_source)
        self.config_files = ("cf", "cf-env")
        create_unique_branch(self.good_source, self.config_files)

        workspace_dir = mkdtemp()
        self.addCleanup(rmtree, workspace_dir)
        self.ws = mojo.Workspace("workspace_name", workspace_dir)

        self.addCleanup(os.chdir, os.getcwd())

    def test_inits_branch(self):
        s = mojo.Spec(self.good_source)
        self.assertIsInstance(s.local_branch, mojo.bzr.BzrBranch)
        self.assertEqual(s.local_branch.location, self.good_source)


class BicommandTest(TestCase):
    "Test that bicommand produces the same output as getstatusoutput"
    def test_bicommand(self):
        cmds = ["ls -1 /tmp",  # rarely will fail, will usually be fast
                "ls ~",  # never (?) will fall, will usually be fast
                "uname",  # shouldn't ever fail, will be fast
                "find /this/file/should/not/exist",  # should be stderr
                "find /usr/share/doc/dpkg",  # will be slow
                ]
        for cmd in cmds:
            bc_status, bc_output = mojo.utils.bicommand(cmd)
            c_status, c_output = commands.getstatusoutput(cmd)

            # We don't care about return codes, we just want to confirm they're
            # both returning zero or non-zero
            if bc_status != 0:
                bc_status = 1
            if c_status != 0:
                c_status = 1

            self.assertEqual(bc_status, c_status)
            self.assertEqual(bc_output, c_output)
