import os
import pwd
import grp
import errno


def mkdirs(path, user=os.environ.get('USER'), group=None, mode=0755):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise OSError("Creation of directory failed: {}".format(path))
    uid = pwd.getpwnam(user).pw_uid
    if group:
        gid = grp.getgrnam(group).gr_gid
    else:
        gid = pwd.getpwnam(user).pw_gid
    os.chown(path, uid, gid)


def mksymlink(source, link_name):
    if os.path.exists(link_name):
        if os.path.islink(link_name):
            if os.path.realpath(source) == os.path.realpath(link_name):
                return True
            else:
                raise OSError("{} link exists but its target {} does not "
                              "match {}".format(link_name,
                                                os.path.realpath(link_name),
                                                os.path.realpath(source)))
        else:
            raise OSError("{} exists but is not a symlink".format(link_name))
    else:
        os.symlink(source, link_name)
        return True


def get_dep_dir(depname):
    deps_dir = os.environ.get("MOJO_DEPS_DIR",
                              os.path.join(os.path.dirname(__file__),
                                           '..', 'deps'))
    dep_dir = os.path.join(os.path.abspath(deps_dir), depname)
    if os.path.isdir(dep_dir):
        return dep_dir
    else:
        return None


def get_dep_prog(progname):
    locations = os.environ.get("PATH").split(os.pathsep)
    for location in locations:
        candidate = os.path.join(location, progname)
        if os.path.isfile(candidate):
            return candidate
    deps_dir = os.environ.get("MOJO_DEPS_DIR",
                              os.path.join(os.path.dirname(__file__),
                                           '..', 'deps'))
    prog = os.path.join(os.path.abspath(deps_dir), progname, progname)
    if os.path.isfile(prog):
        return prog
    else:
        return None


# This meant as a replacement for commands.getstatusoutput.  By default, it
# works more or less identically.  However, if you pass showoutput=True as
# its second parameter, it will show the process of your command as it goes.
# This is useful for long-running commands whose output is interesting to
# watch as well as to programatically analyse once it's finished.
# Stolen mercilessly from https://gist.github.com/maw/1711614 (with thanks)
def bicommand(command, showoutput=False, read_bytes=1):
    import subprocess
    import select
    import sys
    
    if not showoutput in [True, False]:
        raise "showoutput takes a boolean argument only"
    
    pipe = subprocess.Popen(command, shell=True,
                            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    
    output = ""
    while True:
        to_read, _, _ = select.select([pipe.stdout, pipe.stderr], [], [], 0.1)
        if len(to_read) == 0:
            continue
        
        eof = True
        for stream in to_read:
            line = stream.read(read_bytes)
            if line != "":
                eof = False
            output += line
            if showoutput == True:
                sys.stdout.write(line)
                pass
        
        if eof == True:
            break

        pass
    
    if output[-1:] == '\n':
        output = output[:-1]
        pass
    
    status = pipe.wait()
    
    return status, output


class Unprivileged(Exception):
    pass


# Decorator for methods that must be run as root
def needs_root(method):
    def wrapped(*args, **kwargs):
        if os.geteuid() != 0:
            raise Unprivileged("This function must be run as a privileged "
                               "user")
        return method(*args, **kwargs)
    return wrapped
